c * * * * * * * * * * * * * * * * * * * * * * * * *
c    Driver for SROCK2
c    (Use plotsol.m to plot in Matlab the solution)
c * * * * * * * * * * * * * * * * * * * * * * * * *
c
c    This driver shows how to use SROCK2. 
c    It solves the stiff system of SDEs
c    resulting from discretization of the 
c    heat equation with noise:
c     
c    u_t=nu*u_{xx} _ beta u + sigma (u+V_0) \dot W(x,t)  
c        for 0<= t <=1, 0<= x <= 1
c
c    with Neumann boundary conditions
c
c    where \dot W(x,t) is a space-time Stratonovitch noise. 
c--------------------------------------------------------
c ----- to integrate with rock2.f ----- 
      include 'srock2.f' 
			include 'zufall.f'
			
      implicit double precision (a-h,o-z)
c --- parameters for the problem -----
c LTS parameters
      parameter (neqn=200,nsto=200)
c random generator parameters
      CHARACTER(len=32) :: arg
			CHARACTER*60 NAME
			CHARACTER*4 xstring
c			parameter (nexp=10**6)
			parameter (nexp=1)
			parameter (numexp=1)
      double precision TABH(numexp),TABERR(numexp)
c
			double precision svblk(608),winc(nsto)
c To store statistics
      dimension istat(4)
			
c--------------------------------------------------------
c      Work is of length 12*neqn+nsto*neqn (general noise)
c      Work is of length 12*neqn+neqn (diagonal noise)
c      HERE the noise is diagonal
c-------------------------------------------------------- 
c ----- to integrate with rock2.f -----
c      dimension y(neqn),work(12*neqn+nsto*neqn)
      dimension y(neqn),work(12*neqn+neqn)
			dimension yzero(neqn)
      integer iwork(12),idid
      external fheat,gnoise,genrnd


c --- common parameters for the problem -----
c SPDE parameters
      COMMON /PARAM/sigma,anu,amu,vzero
			sigma=2.D-2*SQRT(neqn-1.0D0)
			amu=1.0D0 
			anu=1.d-2
			vzero=10.0D0
			sigma=4.D-3*SQRT(neqn-1.0D0)

c number of time steps
      NSTEPS=50


c ----- file for solution -----
      OPEN(12,FILE='sol.dat')
			
c Initial condition
      api=4.0D0*ATAN(1.0D0)
      DO I=1,neqn
			x=(I-1.0D0)/(neqn-1)
			Yzero(I)=-70.D0 + 20.0D0*COS(5.0D0*api*x)*(1-x)
			END DO
			write (12,*) (Yzero(I),I=1,neqn)
			


c initialize random number generator	
			IF (IARGC().GE.1) THEN
			CALL getarg(1, arg)
      Read( arg, '(i10)' )  iseed
			ELSE
			iseed=0
			END IF
      call zufalli(iseed)
c--------------------------------------------------------
c     Initialise iwork: 
c      iwork(1)=1  RHO returns an upper bound for the spectral radius
c      iwork(2)=1  Case of a diagonal noise
c      iwork(3)=0  Return and solution at tend
c      iwork(4)=0  Force polynomial degree if nonzero (automatic otherwise)
c--------------------------------------------------------
      iwork(1)=1
c nondiagonal noise? (1 for yes)
      iwork(2)=1
c      iwork(3)=0
      iwork(3)=1
      iwork(4)=0
			write (6,*) 'stiffness',4.d0*neqn**2/nsteps*anu
			
c for neqn=200
			iwork(4)=9
			iwork(4)=0
			
c ----- end point of integration -----
      tend=1.d0
c			nloop=numexp/2
c      do iloop=1,2*nloop

      do iloop=1,1
c ----- initial step size -----
		  h=tend/NSTEPS
			TABH(iloop)=h
			t_old=0.0d0
c ----- integration -----
      write(6,*) 'Integration of the SDE problem' 
			write (6,*) 'mdeg:',iwork(4)
			write (6,*) 'nsteps:',nsteps
			write (6,*) 'number of samples',nexp
			write (6,*) 'dimensions:',neqn,nsto
			call cpu_time(time0)
c --------- initial values -------
			t=0.d0
			avg2=0.d0
			stddev=0.d0
      do iexp=1,nexp
      do i=1,neqn
			y(i)=yzero(i)
			end do

			sqh=sqrt(h)
			sqh3=sqrt(3*h)
			
			do istep=1,NSTEPS
			
      call srock2(neqn,nsto,t,1,h,winc,y,fheat,gnoise,genrnd,
     &           work,iwork,idid) 

				if (iexp.EQ.1) write (12,*) (Y(I),I=1,neqn)
		
			if (iexp.eq.1) then
			if (istep.eq.1) then
			write(6,*) 'FIRST STEP'
      write(6,*) 'Max number of stages used=',iwork(10)
      write (6,91) iwork(5),iwork(6),iwork(7),iwork(8)
			do i=1,4
			istat(i)=iwork(i+4)
			end do
			else
			do i=1,4
			istat(i)=istat(i)+iwork(i+4)
			end do
			end if
			end if
			end do
			end do

			call cpu_time(time1)
			write (6,*) 'CPU TIME',time1-time0
			write (6,*) 'nsteps',nsteps

c ----- print statistics -----
      write(6,*) 'Solution is tabuled in file sol.dat'
      write(6,*) 'Max number of stages used=',iwork(10)
      write (6,91) istat(1),istat(2)
 91   format(' Number of f evaluation=',i5,' steps=',i4)
		  end do
			
			
c--------------------------------------------------------
c     End of main program
c--------------------------------------------------------
      end      
c--------------------------------------------------------
c     The subroutine RHO gives an estimation of the spectral 
c     radius of the Jacobian matrix of the problem. This
c     is a bound for the whole interval and thus RHO is called
c     once.
c--------------------------------------------------------
      double precision function rho(neqn,t,y)
      implicit double precision (a-h,o-z)
      COMMON /PARAM/sigma,anu,amu,vzero
			rho=4*neqn**2*anu
      return
      end
c--------------------------------------------------------
c     The subroutine FHEAT compute the value of f(x,y) and
c     has to be declared as external.
c--------------------------------------------------------
      subroutine fheat(neqn,x,y,f)
      implicit double precision (a-h,o-z)
      dimension y(neqn),f(neqn)
      COMMON /PARAM/sigma,anu,amu,vzero
c Neumann condition
			f(1)=2.d0*anu*(Y(2)-Y(1))*(neqn-1)**2
      DO I=2,neqn-1
      f(I)=anu*(Y(I+1)-2.0D0*Y(I)+Y(I-1))*(neqn-1)**2
			END DO
c Neumann condition
			f(neqn)=2.d0*anu*(Y(neqn-1)-Y(neqn))*(neqn-1)**2
			
			DO I=1,neqn
			f(I)=f(I)-amu*Y(I)
			END DO
			return
      end
c--------------------------------------------------------
c     The subroutine GNOISE compute the value of g(x,y) and
c     has to be declared as external.
c--------------------------------------------------------
      subroutine GNOISE(neqn,nsto,isto,x,y,g)
      implicit double precision (a-h,o-z)
      dimension y(neqn),g(nsto)
      COMMON /PARAM/sigma,anu,amu,vzero
			
c ---------- discretisation of diffusion -------
c ------ periodic boundary conditions -------
      do i=1,nsto
       g(i)=(y(i)+vzero)*sigma
			end do
			return
      end
c --------------------------------------------------------
c  ---- generate random numbers needed at each time step
c ----- uses the code zufall by W. P. Petersen (ETH Zurich)
c --------------------------------------------------------
      subroutine GENRND(NSTO,WINC)
			implicit double precision (a-h,o-z)
      double precision winc(2*nsto)
c Gaussian random variables
c      call normalen(nsto,winc)
c Discrete Gaussian path
        call zufall(nsto,winc)
				tmp=sqrt(3.d0)
				DO I=1,nsto
				IF (6.0D0*winc(I).LT.1.0D0) THEN
			  Winc(I)=-tmp
			  ELSE IF (6.0D0*winc(I).LE.2.0D0) THEN
			  Winc(I)=tmp
			  ELSE
			  Winc(I)=0.0D0
			  END IF
				END DO
c
c				DO I=nsto+1,2*nsto
c				IF (winc(I).LT.0.5D0) THEN
c			  Winc(I)=1.d0
c				ELSE
c			  Winc(I)=-1.d0
c			  END IF
c				END DO
			 return
			 end
