c * * * * * * * * * * * * * * * * * * * * * * * * *
c    Driver for SROCK2
c    (Use plotconv.m to plot in Matlab the convergence curve
c    using the average over nexp=1e6 trajectories)
c * * * * * * * * * * * * * * * * * * * * * * * * *
c
c    This driver shows how to use SROCK2. 
c    It solves the stiff system of SDEs
c    dX(t) = X(t)dt + \sum_{j=1}^10 1/a_j (X(t)+1/b_j) dW(t)
c    with a_j , j = 1, . . . , 10 given respectively by 
c    10, 15, 20, 25, 40, 25, 20, 15, 20, 25
c    and with b_j , j = 1, . . . , 10 given respectively by 
c    2, 4, 5, 10, 20, 2, 4, 5, 10, 20. 
c    and compute the second moment error for X(T) at time T=1.
c     
c--------------------------------------------------------
      include 'srock2.f' 
			include 'zufall.f'
			
      implicit double precision (a-h,o-z)
c --- parameters for the problem -----
c LTS parameters
      parameter (neqn=1,nsto=10)
c random generator parameters
      CHARACTER(len=32) :: arg
			CHARACTER*60 NAME
			CHARACTER*4 xstring
c number of samples
			parameter (nexp=10**6)
			parameter (numexp=8)
      double precision TABH(numexp),TABERR(numexp)
c
			double precision svblk(608),winc(2*nsto)
c To store statistics
      dimension istat(4)
c--------------------------------------------------------
c      Work is of length 12*neqn+nsto*neqn
c-------------------------------------------------------- 
c ----- to integrate with srock2.f -----
      dimension y(neqn),work(12*neqn+nsto*neqn)
      integer iwork(13),idid
      external fheat,gnoise,genrnd

c --- common parameters for the problem -----
      common/deltax/deltax,amu,alambda,tau,ibug
c -------- dimensions and initialisations --------       
        amu=1.0d0
c initialize random number generator	
			IF (IARGC().GE.1) THEN
			CALL getarg(1, arg)
      Read( arg, '(i10)' )  iseed
			ELSE
			iseed=0
			END IF
      call zufalli(iseed)
c--------------------------------------------------------
c     Initialise iwork: 
c      iwork(1)=1  RHO returns an upper bound for the spectral radius
c      iwork(2)=1  Case of a diagonal noise
c      iwork(4)=0  Force polynomial degree if nonzero (automatic otherwise)
c--------------------------------------------------------
      iwork(1)=1
      iwork(2)=0
      iwork(4)=0
			
c     polynomial degree			
			iwork(4)=13
			
c ----- end point of integration -----
      tend=1.d0
			nloop=numexp/2
      do iloop=1,2*nloop
c ----- initial step size -----
			NSTEPS=floor(sqrt(2.d0)**(iloop)+0.5d0)
c      NSTEPS=2**iloop
			if (iloop.eq.0) NSTEPS=1
			if (iloop.eq.2*nloop+1) NSTEPS=32
		  h=tend/NSTEPS
			TABH(iloop)=h
			t_old=0.0d0
c ----- integration -----
      write(6,*) 'Integration of the SDE problem' 
			write (6,*) 'mdeg:',iwork(4)
			write (6,*) 'nsteps:',nsteps
			write (6,*) 'number of samples',nexp
			write (6,*) 'dimensions:',neqn,nsto
			call cpu_time(time0)
c --------- initial values -------
			t=0.d0
			avg2=0.d0
			stddev=0.d0
      do iexp=1,nexp
			if (mod(iexp,nexp/10**1).eq.0) 
     &     write (6,*) 'iexp',(iexp*1.d2)/nexp,
     &            '% ',avg2/iexp,iloop
c initial condition
      y(1)=1.d0

      call srock2(neqn,nsto,t,nsteps,h,winc,y,fheat,gnoise,genrnd,
     &           work,iwork,idid) 

c problem 18
      if (iexp.eq.1) avg2ex=(-68013.d0 - 458120.d0*exp(1.d0) 
     &       + 14926133.d0*exp(2.d0))/14400000.d0
		  err=y(1)**2-avg2ex
      avg2=avg2+err
      stddev=stddev+err**2
			end do
			avg2=avg2/nexp
			stddev=stddev/nexp
			call cpu_time(time1)
			write (6,*) 'CPU TIME',time1-time0
			write (6,*) 'nsteps',nsteps,'avg2',avg2

			write (6,*) 'avg2ex',avg2ex,avg2,abs(avg2)/avg2ex,
     &     'dev',sqrt(stddev-avg2**2)
      TABERR(iloop)=avg2
c			write (9,*) h,avg2,nsteps,avg2ex
c ----- print statistics -----
      write(6,*) 'Max number of stages used=',iwork(10)
      write (6,91) istat(1),istat(2)
 91   format(' Number of f evaluation=',i5,' steps=',i4)
		  end do
c
      Write( xstring, '(i4)' )  iseed
			NAME='resavg'//xstring//'.dat'
			if (iseed.eq.0) NAME='dataavg.dat'
			write(6,*) 'SAVING IN FILE ',NAME
      OPEN(9,FILE=NAME)
      REWIND(9)
			Write (9,*) (TABH(iloop),iloop=1,numexp),
     &   (TABERR(iloop),iloop=1,numexp)
			
c--------------------------------------------------------
c     End of main program
c--------------------------------------------------------
      end      
c--------------------------------------------------------
c     The subroutine RHO gives an estimation of the spectral 
c     radius of the Jacobian matrix of the problem. This
c     is a bound for the whole interval and thus RHO is called
c     once.
c--------------------------------------------------------
      double precision function rho(neqn,t,y)
      implicit double precision (a-h,o-z)
      common/deltax/deltax,amu,alambda,tau,ibug
			rho=4.d0*tau + 1.0D0
			rho=10.
      return
      end
c--------------------------------------------------------
c     The subroutine FHEAT compute the value of f(x,y) and
c     has to be declared as external.
c--------------------------------------------------------
      subroutine fheat(neqn,x,y,f)
      implicit double precision (a-h,o-z)
      dimension y(neqn),f(neqn)
      common/deltax/deltax,amu,alambda,tau,ibug
			f(1)=amu**2*y(1)
      return
      end
c--------------------------------------------------------
c     The subroutine GNOISE compute the value of g(x,y) and
c     has to be declared as external.
c--------------------------------------------------------
      subroutine GNOISE(neqn,nsto,isto,x,y,g)
      implicit double precision (a-h,o-z)
      dimension y(neqn),g(neqn)
      common/deltax/deltax,amu,alambda,tau,ibug
			
c ---------- discretisation of diffusion -------
c ------ periodic boundary conditions -------
       if (y(1).lt.1.d0/20.d0) then
         write (6,*) 'Warning in SQRT in GNOISE!!',y(1)
				 g(1)=0.d0
				 pause
				 return
       end if
       if (isto.eq.1) g(1)=amu/10.0d0*sqrt(y(1)+1.d0/2.d0)
       if (isto.eq.2) g(1)=amu/15.0d0*sqrt(y(1)+1.d0/4.d0)
			 if (isto.eq.3) g(1)=amu/20.0d0*sqrt(y(1)+1.d0/5.d0)
			 if (isto.eq.4) g(1)=amu/25.0d0*sqrt(y(1)+1.d0/10.d0)
			 if (isto.eq.5) g(1)=amu/40.0d0*sqrt(y(1)+1.d0/20.d0)
			 if (isto.eq.6) g(1)=amu/25.0d0*sqrt(y(1)+1.d0/2.d0)
			 if (isto.eq.7) g(1)=amu/20.0d0*sqrt(y(1)+1.d0/4.d0)
			 if (isto.eq.8) g(1)=amu/15.0d0*sqrt(y(1)+1.d0/5.d0)
			 if (isto.eq.9) g(1)=amu/20.0d0*sqrt(y(1)+1.d0/10.d0)
			 if (isto.eq.10) g(1)=amu/25.0d0*sqrt(y(1)+1.d0/20.d0)
       return  
      end
c --------------------------------------------------------
c  ---- generate random numbers needed at each time step
c ----- uses the code zufall by W. P. Petersen (ETH Zurich)
c --------------------------------------------------------
      subroutine GENRND(NSTO,WINC)
			implicit double precision (a-h,o-z)
      double precision winc(2*nsto)
c Gaussian random variables
c      call normalen(nsto,winc)
c Discrete Gaussian path
        call zufall(2*nsto,winc)
				tmp=sqrt(3.d0)
				DO I=1,nsto
				IF (6.0D0*winc(I).LT.1.0D0) THEN
			  Winc(I)=-tmp
			  ELSE IF (6.0D0*winc(I).LE.2.0D0) THEN
			  Winc(I)=tmp
			  ELSE
			  Winc(I)=0.0D0
			  END IF
				END DO
c
				DO I=nsto+1,2*nsto
				IF (winc(I).LT.0.5D0) THEN
			  Winc(I)=1.d0
				ELSE
			  Winc(I)=-1.d0
			  END IF
				END DO
			 return
			 end