      subroutine srock2(neqn,nsto,t,nsteps,h,winc,y,f,g,
     &   genrnd,work,iwork,idid)
c ----------------------------------------------------------

c    Second weak order explicit stabilized methods 
c    for stiff It stochastic differential equations 
c    dX(t) = f(t,X(t)) dt + sum_i g^i(t,X(t)) dW_i(t)
c
c    This algorithm is described in
c    [ ] A. Abdulle, G. Vilmart, and K.C. Zygalakis 
c        Second weak order explicit stabilized methods
c        for stiff stochastic differential equations 
c        to appear in SIAM J. Sci. Comput., 2013
c
c    Version of December 12, 2012
c
c    PLEASE CITE THE ABOVE PAPER WHEN USING SROCK2:
c    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c    This code is a modification of the code ROCK2 by A. Abdulle. 

c ----------------------------------------------------------
c    ROCK2 is based on a family of second order explicit 
c    Runge-Kutta methods with nearly optimal stability domain 
c    on the negative real axis. The numerical method is 
c    based on a three-term recursion relation.
c    The size (along the negative axis) of the stability 
c    domains increases quadratically with the stage number.
c
c    Intended for problems of large dimensions with
c    eigenvalues of the Jacobian close to the negative
c    real axis. Typically for problems originating from
c    parabolic PDEs.
c
c    Author: A. Abdulle
c            Universite de Geneve, Dept. de mathematiques
c            Ch-1211 Geneve 24, Switzerland 
c            now at EPFL, Switzerland assyr.abdulle@epfl.ch
c
c    Version of April 2002  minor bug corrected
c    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
c    PLEASE CITE THE FOLLOWING PAPER WHEN USING ROCK2:
c
c    [ ] A. Abdulle & A.A. Medovikov
c        Second order Chebyshev methods based on
c        orthogonal polynomials
c        Numer. Math.  90, no. 1  (2001), 1--18.
c        
c--------------------------------------------------------------------
c     Input parameters  
c     ----------------  
c     NEQN:       Number of differential equations of the system 
c                 (integer).
c
c     NSTO:       Number of independent Wiener processes.
c
c     T:          Initial point of integration (double precision).
c
c     NSTEPS:     Number of timesteps
c
c     H:          Constant timestep size
c
c     Y(NEQN):    Initial value of the solution 
c                 (double precision array of length neqn).
c
c     F:          Name (external) of subroutine computing the value 
c                 of f(x,y). Must have the form
c
c                   subroutine f(neqn,t,y,dy)
c                   double precision y(neqn),dy(neqn)
c                   integer neqn
c                   dy(1)=...
c                   ...
c                   dy(neqn)=...
c                   return
c                   end 
c
c     G:          Name (external) of subroutine computing the value 
c                 of g^i(x,y) where i is the noise index. 
c                 Must have the form
c
c                   subroutine g(neqn,nsto,i,t,y,dy)
c                   double precision y(neqn),dy(neqn)
c                   integer neqn
c                   dy(1)=...
c                   ...
c                   dy(neqn)=...
c                   return
c                   end 
c
c                 Implementation:   
c                 for stability issues when the problem 
c                 is originating from parabolic PDEs, transforming 
c                 inhomogeneous boundary conditions in homogeneous ones
c                 (by adding the appropriate function to the right-hand side)
c                 may increase the performance of the code.
c
c     RHO:        Name (external) of a function (double precision) 
c                 giving the spectral radius of the Jacobian 
c                 matrix  of f at (t,y). Must have the form
c                 
c                   double precision function rho(neqn,t,y)
c                   double precision y(neqn),t
c                   integer neqn
c                   ...
c                   rho=... 
c                   return
c                   end
c               
c                 N.b. Gerschgorin's theorem can be helpful. If the
c                 Jacobian is known to be constant it should be 
c                 specified by setting iwork(2)=1 (see below).
c
c                 ROCK2 can also compute this estimate. In that 
c                 case, provide a dummy function rho(neqn,t,y) and 
c                 set iwork(1)=0 (see below).
c
c                 If it is possible to give an estimate of 
c                 the spectral radius, it should be preferred to
c                 the estimate computed internally by ROCK2.
c
c     IWORK(*):   Integer array of length 12 that gives information 
c                 on how the problem is to be solved and communicates 
c                 statistics about the integration process.
c               
c     IWORK(1):   =0 ROCK2 attempts to compute the spectral radius 
c                    internally. Define a dummy function
c
c                    double precision function rho(neqn,t,y)
c                    double precision y(neqn),t
c                    integer neqn
c                    rho=0.d0 
c                    return
c                    end
c
c                 =1 RHO returns an upper bound of the spectral 
c                    radius  of the Jacobian matrix of f at (t,y).
c    
c     IWORK(2):   =0 General non-commutative noise case                 
c                 =1 Diagonal noise case
c
c     IWORK(4):   =0 automatic choice of stage number
c                 otherwise force stage number to IWORK(4)
c
c     WORK(*) :     Workspace of length 12*neqn+nsto if iwork(2)=0,
c                   otherwise of length 12*neqn+neqn*nsto.
c                   Work(1),..,work(4*neqn) serve as
c                   working space for the solution of
c                   the ode.
c                   Work(4*neqn+1),..,work(5*neqn)
c                   serve as working space for the
c                   internal computation of the 
c                   spectral radius of the Jacobian.
c                  
c     IDID:         Report on successfulness upon return
c                   (integer).
c
c
c     Output parameters 
c     -----------------
c     T:          T-value for which the solution has been computed
c                 (after successful return t=tend).
c
c     Y(NEQN):    Numerical solution at tend.
c
c     IDID:       Reports what happened upon return
c
c     IDID        =1 Successful computation t=tend.
c                 =2 Successful computation of one step
c                    to continue call ROCK2 again without
c                    altering any arguments.
c                 =-1 Invalid input parameters.
c                 =-2 Stepsize becomes to small.
c                 =-3 The method used in ROCK2 to estimate 
c                     the spectral radius did not converge.
c
c     IWORK(5)    =Number of function evaluations.
c     IWORK(6)    =Number of steps.
c     IWORK(7)    =Number of accepted steps.
c     IWORK(8)    =Number of rejected steps.
c     IWORK(9)    =Number of evaluations of f used
c                  to estimate the spectral radius
c                  (equal to zero if iwork(1)=1).
c     IWORK(10)   =Maximum number of stages used.
c     IWORK(11)   =Maximum value of the estimated 
c                  bound for the spectral radius 
c                  (rounded to the nearest integer).
c     IWORK(12)   =Minimum value of the estimated  
c                  bound for the spectral radius 
c                  (rounded to the nearest integer).
c                       
c   
c    Caution:     The variable UROUND (the rounding unit) is set to 
c    -------      1.0d-16 and may depends on the machines.
c-------------------------------------------------------------------
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
c         Numerical method
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c      The nearly optimal stability polynomial is computed
c      as a product:  R_s(z)=P_{s-2}(z)*w(z).
c      We realize this polynomial as a Runge-Kutta method
c      with a three-term recurrence formula for the first
c      s-2 stages (with P_{s-2}(z) as stability polynomial)
c      and a 2-stage finishing procedure (with w(z) as 
c      stability polynomial).
c
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
c         Stability functions and three-term recurrence relation.
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c     
c     The stability functions: R_j(z)=P_{j-2}(z)  (internal j<=ms-2)
c                             R_{ms}=P_{s-2}(z)*w(z) ( absolute) 
c                             w(z)=1+sigma*z+tau*z^2 
c                 
c     P_j(z) orthogonal with respect to w(z)^2/sqrt{1-x^2}
c     
c     Recurrence formula: 
c     
c       P_j(z)=(a_j*z-b_j)*P_{j-1}(z)-c_j*P_{j-2}(z) 
c                   j=1..ms-2 b_1=-1,c_1=0
c
c     Normalization: P_(0)=1  =>b_{j}=-(1+c_{j}) 
c
c     Runge-Kutta formula:
c
c     g_j(z)=a_{j}*f(g_{j-1})-b_{j}*g_{j-1}-c_{j}*g_{j-2} j=1..ms-2 
c
c     Data (rec. param.): rec(i)= a_1,a_2,c_2,a_3,c_3,.,a_(ms-2),
c                          c_(ms-2)  for ms=1,3,5,..  
c
c     The two-stage finishing procedure:
c
c     g_{s-1}=g_{s-2} + h*sigma*f(g_{s-2}
c     ge_{s}=g_{s-1} + h*sigma*f(g_{s-1}
c     g_{s}=ge_{s} -h*sigma*(1-(sigma/tau)*(f(g_{s-1}-f(g_{s-2})
c
c     Embedded method: ge_{s}
c
c     Datas (finish. proced.): fp1(ms) =sigma  
c                              fp2(ms) =-sigma(1-(sigma/tau) 
c                              ms=1,3,5,.. 
c
c     Chosen degrees: s=3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,
c     20,21,22,24,26,28,30,32,35,38,41,45,49,53,58,63,68,74,80,87,95,
c     104,114,125,137,150,165,182,200  ms=s-2
c
c------------------------------------------------------------------
c
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***          
c             Declarations 
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c
      double precision y(neqn),work(*),winc(*),
     & t,tend,h,uround,
     & recf(4476),fp1(46),fp2(46),  
     & recf2(184),recalph(46)
      integer iwork(12),ms(46),neqn,i,n1,n2,n3,n4,n5,n6,n7,n8,
     & n9,n10,n11,n12,n13,nsto,idid,nsteps
      logical arret
      external f,g,genrnd
c -------- Uround: smallest number satisfying 1.d0+uround>1.d0
      data uround/1.0d-16/ 
c             
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c             Data of the stability polynomials
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
c Gilles modif: see included file 'rectp.f'
      include 'rectp.f'

c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
c             Initializations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
      arret=.false.
c -------- Prepare the entry-points for the arrays in work.--------     
      n1=1
      n2=n1+neqn
      n3=n2+neqn
      n4=n3+neqn
			n5=n4+neqn
			n6=n5+neqn
			n7=n6+neqn
			n8=n7+neqn
			n9=n8+neqn
			n10=n9+neqn
			n11=n10+neqn
			n12=n11+neqn
			n13=n12+neqn
c -------- Call to the core integrator. -----------
      call rockcore(neqn,nsto,t,nsteps,h,winc,y,f,g,work,work(n1),
     & work(n2),work(n3),work(n4),ms,recf,fp1,
     & fp2,iwork,arret,uround,idid,work(n5),work(n6),recf2,recalph,
     & work(n7),work(n8),work(n9),work(n10),work(n11),
     & work(n13))
      return
      end
c ----------------------------------------------
c     End of subroutine ROCK2.
c ----------------------------------------------
c
      subroutine rockcore(neqn,nsto,t,nsteps,h,winc,y,f,g,work,
     & yn,fn,yjm1,yjm2,ms,recf,fp1,fp2,iwork,
     & arret,uround,idid,yks1,yks2,recf2,recalph,yjm3,yjm4,y2,fnc,yrk0,
     & stomat)
c ----------------------------------------------
c    Core integrator for ROCK2.
c ---------------------------------------------- 
c             Declarations
c ----------------------------------------------
       double precision y(*),yn(*),fn(*),work(*),yjm1(neqn),
     & yjm2(neqn),recf(4476),fp1(46),fp2(46),
     & tend,h,hnew,hp,facmax,fac,facp,rho,uround,
     & eigmax,t,told,alpha,beta,
     & recf2(184),recalph(46),yks1(neqn),yks2(neqn),yjm3(neqn),
     & yjm4(neqn),fnc(neqn),y2(neqn),yrk0(neqn),
     & stomat(*),winc(*)
       integer ms(46),mp(2),iwork(12),neqn,nsto,mdeg,i,nsteps,istep,
     & nrho,mdego,nrej,idid
       logical last,reject,arret
       external f,g
			 double precision cks,tmp,rnd
			 integer j,k
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
c             Initializations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
      iwork(5)=0
      iwork(6)=0
      iwork(7)=0
      iwork(8)=0
      iwork(9)=0
      iwork(10)=0      
      iwork(11)=0
      iwork(12)=0
      told=0.d0
      hp=h
      last=.false.
      reject=.false. 
      nrho=0
      idid=1
c			mdego=mdeg
			mdego=0
c -------- Big loop
      do istep=1,nsteps
				tmp=sqrt(h)
				CALL GENRND(NSTO,WINC)
				DO I=1,NSTO
				WINC(I)=WINC(I)*tmp
				END DO
c non diagonal noise
				if (iwork(2).eq.0) then
				DO I=NSTO+1,2*NSTO
				WINC(I)=WINC(I)*tmp
				END DO
				end if
c -------- Initialization of the integration step.--------   
10    do i=1,neqn
       yn(i)=y(i)
      end do
      call f(neqn,t,yn,fn)
      iwork(5)=iwork(5)+1
c -------- Spectral radius.--------
      if (iwork(4).gt.0) then
c force the degree 
        mdeg=iwork(4)
				if (mdeg.gt.200) 
     &      write (6,*) 'Warning: stages>200'
			else
      if (nrho.eq.0) then
c        if (iwork(6).eq.0.or.iwork(2).eq.0) then
        if (iwork(6).eq.0) then				
c ------- Computed externally by rho.--------
          if (iwork(1).eq.1) then
            eigmax=rho(neqn,t,yn)
c SROCK: multiply by two the spectral radius
						eigmax=2.d0*eigmax
c
            if (idnint(eigmax).gt.iwork(11)) 
     &         iwork(11)=idnint(eigmax)
            if (iwork(6).eq.0) iwork(12)=iwork(11)
            if (idnint(eigmax).lt.iwork(12)) 
     &         iwork(12)=idnint(eigmax)
c ------- Computed internally by rocktrho.--------  
                             else
            call rocktrho(neqn,t,y,f,yn,fn,work,yjm1,
     &                   yjm2,eigmax,uround,idid,iwork)
		        eigmax=0.5d0*eigmax
            if (idnint(eigmax).gt.iwork(11)) 
     &         iwork(11)=idnint(eigmax)
            if (iwork(6).eq.0) iwork(12)=iwork(11)
            if (idnint(eigmax).lt.iwork(12)) 
     &         iwork(12)=idnint(eigmax)
          end if
        end if
      end if
c -------- The number of stages.--------
      mdeg=sqrt((1.5d0+h*eigmax)/0.811d0)+1
      if (mdeg.gt.200) then
        h=0.8d0*(200.d0**2*0.811d0-1.5d0)/eigmax
        mdeg=200
        last=.false.
				write (6,*) 'Warning: stepsize h too large (stages>200)'
      end if
			end if
			
      mdeg=max(mdeg,3)-2
      if (mdeg.ne.mdego) then
        call mdegr(mdeg,mp,ms)
      end if
      if (mdeg+2.gt.iwork(10)) iwork(10)=mdeg+2
c -------- Computation of an integration step.--------
c ROCK2 method
      if (iwork(2).eq.0) then
      call rtstep(neqn,nsto,t,h,y,f,g,yn,fn,yjm1,yjm2,mdeg,mp,
     & recf,fp1,fp2,recf2,recalph,yks1,yks2,yjm3,stomat,winc)
		  else
      call rtstepdiag(neqn,nsto,t,h,y,f,g,yn,fn,yjm1,yjm2,mdeg,mp,
     & recf,fp1,fp2,recf2,recalph,yks1,yks2,yjm3,stomat,winc)
			end if
		
c ---
      iwork(6)=iwork(6)+1 
      iwork(5)=iwork(5)+mdeg+1      
        iwork(7)=iwork(7)+1
        t=t+h      
			end do
			idid=1
      return      
      end
c ----------------------------------------------
c     End of subroutine rockcore.
c ----------------------------------------------  
c           
      subroutine rtstep(neqn,nsto,t,h,y,f,g,yn,fn,yjm1,yjm2,mdeg,
     & mp,recf,fp1,fp2,
     & recf2,recalph,yks1,yks2,yjm3,stomat,winc)
c ----------------------------------------------
c  Solut. at t+h by an explicit (mdeg+2)-stages formula.
c-----------------------------------------------
c             Declarations
c-----------------------------------------------
       double precision y(neqn),yn(neqn),fn(neqn),yjm1(neqn),
     & yjm2(neqn),recf(4476),fp1(46),fp2(46),
     & t,h,ci1,ci2,ci3,temp1,temp2,temp3,ato,rto,
     & recf2(184),recalph(46),yks1(neqn),yks2(neqn),yjm3(neqn),
     & stodouble,
     & stomat(nsto,neqn),winc(2*nsto),
     & alpha,ha,beta1,beta2,rnd,tmp
       integer mp(2),neqn,nsto,mdeg,mr,mz,i,j,k
       external f,g
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
c             Initialisations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***
      mz=mp(1)
      mr=mp(2)
			alpha=recalph(mz)
			ha=h*alpha
c -------- First stage.--------
      temp1=ha*recf(mr)
      ci1=t+temp1
      ci2=t+temp1
      ci3=t
      do i=1,neqn
        yjm2(i)=yn(i)
        yjm1(i)=yn(i)+temp1*fn(i)
        if (mdeg.lt.2) then
          y(i)=yjm1(i)
        end if
      end do
c --------- Stages for j=2..mdeg.--------
      do i=2,mdeg
        temp1=ha*recf(mr+2*(i-2)+1)
        temp3=-recf(mr+2*(i-2)+2)
        temp2=1.d0-temp3
        call f(neqn,ci1,yjm1,y)
        ci1=temp1+temp2*ci2+temp3*ci3
        do j=1,neqn
          y(j)=temp1*y(j)+temp2*yjm1(j)+temp3*yjm2(j)
c -------- Shift the value "y" for the next stage.--------
          if (i.lt.mdeg) then
            yjm2(j)=yjm1(j)
            yjm1(j)=y(j)
          end if
        end do
        ci3=ci2
        ci2=ci1
      end do
c --- SROCK2: compute two additional orthog. polynomials.
      temp1=ha*recf2((mz-1)*4+1)
			temp3=-recf2((mz-1)*4+2)
			temp2=1.d0-temp3
			call f(neqn,ci1,y,yks1)
			beta1=temp1+temp2*ci2+temp3*ci3
      do j=1,neqn
				yjm2(j)=yjm1(j)
				yjm1(j)=y(j)
			  yks1(j)=temp1*yks1(j)+temp2*yjm1(j)+temp3*yjm2(j)
			  yjm2(j)=yjm1(j)
				yjm1(j)=yks1(j)
			end do
      temp1=ha*recf2((mz-1)*4+3)
			temp3=-recf2((mz-1)*4+4)
			temp2=1.d0-temp3
			call f(neqn,beta1,yjm1,yks2)
			beta2=temp1+temp2*beta1+temp3*ci2
      do j=1,neqn
			  yks2(j)=temp1*yks2(j)+temp2*yjm1(j)+temp3*yjm2(j)
			end do
c -------- The two-stage finishing procedure.--------
c 2hsigma_alpha
      temp1=2.d0*h*(fp1(mz)*alpha+0.5d0*(1.0d0-alpha))
c 2htau_alpha
      temp2=2.d0*h*(0.5d0*(1.d0-alpha)**2+2.d0*alpha*(1-alpha)*fp1(mz)
     &      +alpha**2*(fp1(mz)*(fp1(mz)+fp2(mz))))
      call f(neqn,ci1,y,yjm2)
c  stage k_{s-1}^*
      do j=1,neqn
        yjm1(j)=y(j)+temp2*yjm2(j)
      end do
   		do j=1,nsto
			call g(neqn,nsto,j,beta2,yks2,stomat(j,1))
			do i=1,neqn
			yjm1(i)=yjm1(i)+stomat(j,i)*winc(j)
			end do
			end do
c	final stage
c deterministic terms
      ci1=ci1+temp2
      call f(neqn,ci1,yjm1,yjm3)
      do j=1,neqn
        y(j)=y(j)+(temp1-0.5d0*h)*yjm2(j)+0.5d0*h*yjm3(j)
      end do
c Milstein term
      do k=1,nsto
			do i=1,neqn
			yjm1(i)=yks2(i)
			yjm2(i)=yks2(i)
			do l=1,nsto
c compute stodouble
      if (l.lt.k) tmp=winc(nsto+l)*abs(winc(nsto+l))
			if (l.eq.k) tmp=h
			if (l.gt.k) tmp=-winc(nsto+k)*abs(winc(nsto+k))
			stodouble=0.5d0*(winc(l)*winc(k)-tmp)
c
			tmp=stomat(l,i)*stodouble
			yjm1(i)=yjm1(i)+tmp
			yjm2(i)=yjm2(i)-tmp
			end do
			end do
			call g(neqn,nsto,k,beta2,yjm1,yjm3)
			do i=1,neqn
			y(i)=y(i)+yjm3(i)*0.5d0
			end do
			call g(neqn,nsto,k,beta2,yjm2,yjm3)
			do i=1,neqn
			y(i)=y(i)-yjm3(i)*0.5d0
			end do
			end do
			
c --- nonlinear term
c compute g(K_s)Delta W_n
			do i=1,neqn
			yjm1(i)=0.d0
			do j=1,nsto
			yjm1(i)=yjm1(i)+stomat(j,i)*winc(nsto+j)
			end do
			end do
			
			
			do i=1,neqn
			yjm2(i)=yks1(i)+yjm1(i)*sqrt(0.5d0)
			yjm3(i)=yks1(i)-yjm1(i)*sqrt(0.5d0)
			end do		
			do j=1,nsto
			call g(neqn,nsto,j,beta2,yjm2,yks1)
			do i=1,neqn
			y(i)=y(i)+yks1(i)*winc(j)*0.5d0
			end do
			call g(neqn,nsto,j,beta2,yjm3,yks1)
			do i=1,neqn
			y(i)=y(i)+yks1(i)*winc(j)*0.5d0
			end do
			end do
      return 
      end
c ----------------------------------------------
c     End of subroutine rtstep.
c ---------------------------------------------- 
c      
c ------ CASE OF DIAGONAL NOISE -----------
c           
      subroutine rtstepdiag(neqn,nsto,t,h,y,f,g,yn,fn,yjm1,yjm2,mdeg,
     & mp,recf,fp1,fp2,
     & recf2,recalph,yks1,yks2,yjm3,stomat,winc)
c ----------------------------------------------
c  Solut. at t+h by an explicit (mdeg+2)-stages formula.
c-----------------------------------------------
c             Declarations
c-----------------------------------------------
       double precision y(neqn),yn(neqn),fn(neqn),yjm1(neqn),
     & yjm2(neqn),recf(4476),fp1(46),fp2(46),
     & t,h,ci1,ci2,ci3,temp1,temp2,temp3,ato,rto,
     & recf2(184),recalph(46),yks1(neqn),yks2(neqn),yjm3(neqn),
     & stodouble,
     & stomat(nsto),winc(nsto),
     & alpha,ha,beta1,beta2,rnd,tmp
       integer mp(2),neqn,nsto,mdeg,mr,mz,i,j,k
       external f,g
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
c             Initialisations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***)
      mz=mp(1)
      mr=mp(2)
			alpha=recalph(mz)
			ha=h*alpha
c -------- First stage.--------
      temp1=ha*recf(mr)
      ci1=t+temp1
      ci2=t+temp1
      ci3=t
      do i=1,neqn
        yjm2(i)=yn(i)
        yjm1(i)=yn(i)+temp1*fn(i)
        if (mdeg.lt.2) then
          y(i)=yjm1(i)
        end if
      end do
c --------- Stages for j=2..mdeg.--------
      do i=2,mdeg
        temp1=ha*recf(mr+2*(i-2)+1)
        temp3=-recf(mr+2*(i-2)+2)
        temp2=1.d0-temp3
        call f(neqn,ci1,yjm1,y)
        ci1=temp1+temp2*ci2+temp3*ci3
        do j=1,neqn
          y(j)=temp1*y(j)+temp2*yjm1(j)+temp3*yjm2(j)
c -------- Shift the value "y" for the next stage.--------
          if (i.lt.mdeg) then
            yjm2(j)=yjm1(j)
            yjm1(j)=y(j)
          end if
        end do
        ci3=ci2
        ci2=ci1
      end do
c --- SROCK2: compute two additional orthog. polynomials.
      temp1=ha*recf2((mz-3)*4+1)
			temp3=-recf2((mz-3)*4+2)
			temp2=1.d0-temp3
			call f(neqn,ci1,y,yks1)
			beta1=temp1+temp2*ci2+temp3*ci3
      do j=1,neqn
				yjm2(j)=yjm1(j)
				yjm1(j)=y(j)
			  yks1(j)=temp1*yks1(j)+temp2*yjm1(j)+temp3*yjm2(j)
			  yjm2(j)=yjm1(j)
				yjm1(j)=yks1(j)
			end do
      temp1=ha*recf2((mz-3)*4+3)
			temp3=-recf2((mz-3)*4+4)
			temp2=1.d0-temp3
			call f(neqn,beta1,yjm1,yks2)
			beta2=temp1+temp2*beta1+temp3*ci2
      do j=1,neqn
			  yks2(j)=temp1*yks2(j)+temp2*yjm1(j)+temp3*yjm2(j)
			end do
c -------- The two-stage finishing procedure.--------
c 2hsigma_alpha
      temp1=2.d0*h*(fp1(mz)*alpha+0.5d0*(1.0d0-alpha))
c 2htau_alpha
      temp2=2.d0*h*(0.5d0*(1.d0-alpha)**2+2.d0*alpha*(1-alpha)*fp1(mz)
     &      +alpha**2*(fp1(mz)*(fp1(mz)+fp2(mz))))
      call f(neqn,ci1,y,yjm2)
c  stage k_{s-1}^*
      do j=1,neqn
        yjm1(j)=y(j)+temp2*yjm2(j)
      end do
			call g(neqn,nsto,0,beta2,yks2,stomat)
			do i=1,neqn
			yjm1(i)=yjm1(i)+stomat(i)*winc(i)
			end do
c	final stage
c deterministic terms
      ci1=ci1+temp2
      call f(neqn,ci1,yjm1,yjm3)
      do j=1,neqn
        y(j)=y(j)+(temp1-0.5d0*h)*yjm2(j)+0.5d0*h*yjm3(j)
      end do
c Milstein term (diagonal noise case)
      do i=1,neqn
			yjm1(i)=yks2(i)
			yjm2(i)=yks2(i)
      end do
      do i=1,nsto
			stodouble=0.5d0*(winc(i)**2-h)
			yjm1(i)=yjm1(i)+stomat(i)*stodouble
			yjm2(i)=yjm2(i)-stomat(i)*stodouble
      end do
      call g(neqn,nsto,0,beta2,yjm1,yjm3)
			do i=1,nsto
			y(i)=y(i)+yjm3(i)*0.5d0
			end do
			call g(neqn,nsto,0,beta2,yjm2,yjm3)
      do i=1,nsto
			y(i)=y(i)-yjm3(i)*0.5d0
			end do
			
c --- nonlinear term (diagonal noise case)
      tmp=sqrt(h)
      do i=1,nsto
			yjm1(i)=yks1(i)+stomat(i)*tmp*sqrt(0.5d0)
			yjm2(i)=yks1(i)-stomat(i)*tmp*sqrt(0.5d0)
      end do
      call g(neqn,nsto,0,beta2,yjm1,yjm3)
			do i=1,nsto
			y(i)=y(i)+yjm3(i)*0.5d0*winc(i)
			end do
			call g(neqn,nsto,0,beta2,yjm2,yjm3)
			do i=1,nsto
			y(i)=y(i)+yjm3(i)*0.5d0*winc(i)
			end do
      return 
      end
c ----------------------------------------------
c     End of subroutine rtstep.
c ---------------------------------------------- 
c      
  

      subroutine mdegr(mdeg,mp,ms)
c-------------------------------------------------------------          
c       Find the optimal degree.
c       MP(1): pointer which select the degree in ms(i)\1,2,..
c             such that mdeg<=ms(i).
c       MP(2): pointer which gives the corresponding position
c       of a_1 in the data recf for the selected degree.
c-------------------------------------------------------------        
c ---------------------------------------------- 
c             Declarations
c ---------------------------------------------- 
      integer ms(46),mp(2),mdeg,i
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
c             Initialisations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
      mp(2)=1
c -------- Find the degree.--------     
      do i=1,46
        if ((ms(i)/mdeg).ge.1) then
          mdeg=ms(i)
          mp(1)=i
          return
        end if
      mp(2)=mp(2)+ms(i)*2-1
      end do
      return
      end  
c ----------------------------------------------
c     End of subroutine mdegr.
c ----------------------------------------------
c   
      subroutine rocktrho(neqn,t,y,f,yn,fn,work,z,fz,eigmax,
     &                   uround,idid,iwork)
c------------------------------------------------------------ 
c     Rocktrho compute eigmax, a close upper bound of the
c     spectral radius of the Jacobian matrix using a 
c     power method (J.N. Franklin (matrix theory)). 
c     The algorithm used is a small change (initial vector
c     and stopping criteria) of that of
c     Sommeijer-Shampine-Verwer, implemented in RKC.
c-------------------------------------------------------------
c             Declarations
c-------------------------------------------------------------
       double precision y(neqn),yn(neqn),fn(neqn),z(neqn),
     & work(*),fz(neqn),t,eigmax,eigmaxo,sqrtu,uround,znor,
     & ynor,quot,dzyn,dfzfn,safe
       integer iwork(12),neqn,n5,i,iter,maxiter,nind,ntest,
     & ind,idid  
       parameter (maxiter=50)
       parameter (safe=1.2d0)
       external f
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** *** 
c             Initialisations
c *** *** *** *** *** *** *** *** *** *** *** *** *** *** ***     
      sqrtu=sqrt(uround)
      ynor=0.d0
      znor=0.d0
      n5=4*neqn
c ------ The initial vectors for the power method are yn --------
c       and yn+c*f(v_n), where vn=f(yn) a perturbation of yn 
c       (if iwork(6)=0) or a perturbation of the last computed
c       eigenvector (if iwork(6).neq.0). 
c
      if (iwork(6).eq.0) then
        do i=1,neqn
          fz(i)=fn(i)
        end do
        call f(neqn,t,fz,z)
        iwork(9)=iwork(9)+1
                         else
        do i=1,neqn
          z(i)=work(n5+i)
        end do
      end if
c ------ Perturbation.--------
      do i=1,neqn
        ynor=ynor+yn(i)**2
        znor=znor+z(i)**2
      end do
      ynor=sqrt(ynor)
      znor=sqrt(znor)
c ------ Normalization of the vector z so that --------
c        the difference z-yn lie in a circle 
c        around yen (ice has a constant modules).
c
      if (ynor.ne.0.d0.and.znor.ne.0.d0) then
        dzyn=ynor*sqrtu
        quot=dzyn/znor
        do i=1,neqn
          z(i)=yn(i)+z(i)*quot
        end do
      elseif(ynor.ne.0.d0) then
        dzyn=ynor*sqrtu
        do i=1,neqn
          z(i)=yn(i)+yn(i)*sqrtu
        end do
      elseif(znor.ne.0.d0) then
          dzyn=uround
          quot=dzyn/znor
          do i=1,neqn
            z(i)=z(i)*quot
          end do
      else
        dzyn=uround
        do i=1,neqn
          z(i)=dzyn
        end do
      end if
c ------ Start the power method.--------
      eigmax=0.d0
      do iter=1,maxiter
        call f(neqn,t,z,fz)
        iwork(9)=iwork(9)+1
        dfzfn=0.d0
        do i=1,neqn
          dfzfn=dfzfn+(fz(i)-fn(i))**2
        end do
        dfzfn=dsqrt(dfzfn)
        eigmaxo=eigmax
        eigmax=dfzfn/dzyn
        eigmax=safe*eigmax
c ------ The stopping criteria is based on a 
c        relative error between two successive
c        estimation ``eigmax'' of the spectral 
c        radius.
c 
        if (iter.ge.2.and.dabs(eigmax-eigmaxo)
     &    .le.(eigmax*0.05d0)) then
c ----- The last eigenvector is stored.--------
          do i=1,neqn
            work(n5+i)=z(i)-yn(i)
          end do
          return
        end if
c ----- The next z is defined by --------
c       z_new=yn+coef*(fz-fn) where
c       coef is chosen so that
c       norm(z_new-yn)=norm(z_old-yn).
c
        if (dfzfn.ne.0.d0) then
          quot=dzyn/dfzfn
          do i=1,neqn
            z(i)=yn(i)+(fz(i)-fn(i))*quot
          end do
        else
c ----- The new z is defined by an arbitrary --------
c       perturbation of the current approximation
c       of the eigenvector.
c
          nind=neqn
          ntest=0
          ind=1+mod(iter,nind)
          if (z(ind).ne.yn(ind).or.ntest.eq.10) then
            z(ind)=yn(ind)-(z(ind)-yn(ind))
          else 
            nind=neqn+ind
            ntest=ntest+1
          end if
        end if
      end do
      write(6,*) 'convergence failure in the 
     & spectral radius computation'
      idid=-3
      return
      end
      
         
        
        
        
           
         
         
       
                        
                  
       
       
           
         
      
      
     


